from django.contrib.auth import authenticate


class FirebaseTokenMiddleware:

    def resolve(self, next, root, info, **kwargs):
        context = info.context

        user = authenticate(request=context, **kwargs)

        if user is not None:
            context.user = user

        return next(root, info, **kwargs)



from django import forms
from django.contrib.auth import get_user_model
from firebase_admin import auth


User = get_user_model()


class UserRegistrationForm(forms.ModelForm):

    class Meta:
        model = User
        fields = ['firebase_uid', User.USERNAME_FIELD] \
            + list(User.REQUIRED_FIELDS)

    def clean_firebase_uid(self):
        firebase_uid = self.data['firebase_uid']

        try:
            # Ensure the specified user exists in Firebase
            user = auth.get_user(firebase_uid)
            self.data['email'] = user.email
            self.data['username'] = user.display_name
        except ValueError:
            raise forms.ValidationError('The firebase_uid is invalid.')

        return firebase_uid



import os

import environ


BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Configure env vars

env = environ.Env(
    DEBUG=(bool, False),
)
env_file_path = os.path.join(BASE_DIR, '.env')
environ.Env.read_env(env_file=env_file_path)


SECRET_KEY = env('SECRET_KEY')

DEBUG = env('DEBUG')

ALLOWED_HOSTS = env.list('ALLOWED_HOSTS')


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'firebase_auth.apps.FirebaseAuthConfig',
    'package_tests.apps.PackageTestsConfig',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'main.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

AUTH_USER_MODEL = 'package_tests.User'

AUTHENTICATION_BACKENDS = ('firebase_auth.authentication.FirebaseAuthentication',)

WSGI_APPLICATION = 'main.wsgi.application'


# Database
# https://docs.djangoproject.com/en/2.1/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}


# Password validation
# https://docs.djangoproject.com/en/2.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.1/topics/i18n/

LANGUAGE_CODE = 'en-au'

TIME_ZONE = 'Australia/Adelaide'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.1/howto/static-files/

STATIC_URL = '/static/'


# Firebase config

GOOGLE_APPLICATION_CREDENTIALS = os.path.join(
    BASE_DIR,
    env('GOOGLE_APPLICATION_CREDENTIALS'),
)



from django.test import TestCase

from firebase_auth.forms import UserRegistrationForm
from package_tests.tests.helpers import (
    stub_email, stub_firebase_uid, setup_mocks, stub_username,
)


setup_mocks()


class UserRegistrationFormTests(TestCase):

    def test_fields(self):
        assert UserRegistrationForm.Meta.fields == [
            'firebase_uid', 'username', 'email',
        ]

    def test_valid_data(self):
        firebase_uid = stub_firebase_uid
        form = UserRegistrationForm(data={
            'firebase_uid': firebase_uid,
        })
        assert form.is_valid()

        saved_user = form.save()
        assert saved_user.firebase_uid == firebase_uid
        assert saved_user.email == stub_email
        assert saved_user.username == stub_username

    def test_invalid_firebase_uid(self):
        form = UserRegistrationForm(data={
            'firebase_uid': 'invalid_firebase_uid',
        })
        assert not form.is_valid()



# pylint: disable=W0212
from django.test import TestCase
from django.test.client import RequestFactory

from firebase_auth.authentication import FirebaseAuthentication
from package_tests.tests.helpers import (
    create_user, setup_mocks, stub_firebase_uid, stub_firebase_token,
)
from package_tests.models import User


setup_mocks()


class FirebaseAuthenticationTests(TestCase):

    def setUp(self):
        self.request_factory = RequestFactory()

    def test_get_auth_token_with_no_token_returns_none(self):
        request = self.request_factory.get('/')
        token = FirebaseAuthentication()._get_auth_token(request)
        assert token is None

    def test_get_auth_token_with_empty_token_returns_none(self):
        request = self.request_factory.get(
            '/',
            HTTP_AUTHORIZATION='',
        )
        token = FirebaseAuthentication()._get_auth_token(request)
        assert token is None

    def test_get_auth_token_with_invalid_token_returns_none(self):
        request = self.request_factory.get(
            '/',
            HTTP_AUTHORIZATION='invalid',
        )
        token = FirebaseAuthentication()._get_auth_token(request)
        assert token is None

    def test_get_auth_token_with_valid_token_returns_user(self):
        request = self.request_factory.get(
            '/',
            HTTP_AUTHORIZATION=stub_firebase_token,
        )
        token = FirebaseAuthentication()._get_auth_token(request)
        assert token == {
            'uid': stub_firebase_uid,
        }

    def test_register_unregistered_user_with_empty_uid_returns_none(self):
        received_user = FirebaseAuthentication()._register_unregistered_user('')
        assert User.objects.count() == 0
        assert received_user is None

    def test_register_unregistered_user_with_invalid_uid_returns_none(self):
        received_user = FirebaseAuthentication()._register_unregistered_user('invalid_uid')
        assert User.objects.count() == 0
        assert received_user is None

    def test_register_unregistered_user_with_valid_uid_returns_user(self):
        assert User.objects.count() == 0

        uid = stub_firebase_uid
        received_user = FirebaseAuthentication()._register_unregistered_user(uid)
        assert User.objects.count() == 1

        expected_user = User.objects.get(firebase_uid=uid)
        assert received_user == expected_user

    def test_get_user_from_token_with_empty_token_returns_none(self):
        token = {
            'uid': '',
        }
        received_user = FirebaseAuthentication()._get_user_from_token(token)
        assert User.objects.count() == 0
        assert received_user is None

    def test_get_user_from_token_with_invalid_token_returns_none(self):
        token = {
            'uid': 'invalid',
        }
        received_user = FirebaseAuthentication()._get_user_from_token(token)
        assert User.objects.count() == 0
        assert received_user is None

    def test_get_user_from_token_with_valid_token_returns_user(self):
        token = {
            'uid': stub_firebase_uid,
        }
        assert User.objects.count() == 0
        received_user = FirebaseAuthentication()._get_user_from_token(token)
        expected_user = User.objects.get(firebase_uid=token['uid'])
        assert User.objects.count() == 1
        assert received_user == expected_user

    def test_authenticate_with_no_token_returns_none(self):
        request = self.request_factory.get('/')
        user = FirebaseAuthentication().authenticate(request)
        assert user is None

    def test_authenticate_with_empty_token_returns_none(self):
        request = self.request_factory.get('/', HTTP_AUTHORIZATION='')
        user = FirebaseAuthentication().authenticate(request)
        assert user is None

    def test_authenticate_with_invalid_token_returns_none(self):
        request = self.request_factory.get('/', HTTP_AUTHORIZATION='invalid')
        user = FirebaseAuthentication().authenticate(request)
        assert user is None

    def test_authenticate_with_valid_token_and_no_user_returns_user(self):
        request = self.request_factory.get(
            '/',
            HTTP_AUTHORIZATION=stub_firebase_token,
        )
        assert User.objects.count() == 0
        received_user = FirebaseAuthentication().authenticate(request)
        assert User.objects.count() == 1
        expected_user = User.objects.get(firebase_uid=stub_firebase_uid)
        assert received_user == expected_user

    def test_authenticate_with_valid_token_and_user_returns_user(self):
        request = self.request_factory.get(
            '/',
            HTTP_AUTHORIZATION=stub_firebase_token,
        )
        expected_user = create_user(firebase_uid=stub_firebase_uid)
        assert User.objects.count() == 1
        authenticated_user = FirebaseAuthentication().authenticate(request)
        assert User.objects.count() == 1
        assert authenticated_user == expected_user

    def test_get_user_with_no_user_returns_none(self):
        user = FirebaseAuthentication().get_user(user_pk=None)
        assert user is None

    def test_get_user_with_valid_pk_returns_user(self):
        expected_user = create_user()
        user = FirebaseAuthentication().get_user(user_pk=expected_user.pk)
        assert user.pk == expected_user.pk

    def test_get_user_with_invalid_pk_returns_none(self):
        user = FirebaseAuthentication().get_user(user_pk=3)
        assert user is None



import importlib
from django.contrib.auth import get_user_model
from django.conf import settings
from firebase_admin import auth

from firebase_auth.apps import firebase_app
from firebase_auth.forms import UserRegistrationForm

User = get_user_model()


class FirebaseAuthentication:

    def _get_auth_token(self, request):
        authorization = request.META.get('HTTP_AUTHORIZATION')
        if authorization:
            encoded_token = authorization.replace('jwt ', '')
        else:
            return None
        decoded_token = None

        try:
            decoded_token = auth.verify_id_token(encoded_token, firebase_app, False)
        except ValueError:
            pass
        except auth.InvalidIdTokenError:
            pass
        except auth.ExpiredIdTokenError:
            pass
        except auth.RevokedIdTokenError:
            pass
        return decoded_token

    def _register_unregistered_user(self, firebase_uid):
        user = None
        form = UserRegistrationForm(data={
            'firebase_uid': firebase_uid,
        })

        if form.is_valid():
            user = form.save()
        errors = form.errors
        return user

    def _get_user_from_token(self, decoded_token):
        firebase_uid = decoded_token.get('uid')
        user = None

        try:
            user = User.objects.get(firebase_uid=firebase_uid)
        except User.DoesNotExist:
            user = self._register_unregistered_user(firebase_uid)
            if hasattr(settings, 'REGISTER_FIREBASE_USER'):
                try:
                    module = settings.REGISTER_FIREBASE_USER.split(".")

                    m = importlib.import_module(".".join(module[0: -1]))
                    register_user = getattr(m, module[-1])
                    register_user(user, decoded_token)
                except Exception as e:
                    pass
        return user

    def authenticate(self, request, **kwargs):
        user = None
        decoded_token = self._get_auth_token(request)

        if decoded_token:
            user = self._get_user_from_token(decoded_token)
        return user

    def get_user(self, user_pk):
        try:
            user = User.objects.get(pk=user_pk)
        except User.DoesNotExist:
            user = None
        return user
